/* Sample code Texture 6 by Rudy Rucker, CS Dept, SJSU, May 18, 2003.
This code is public domain and can be freely reused. Based on the Molecules 6
build. */

#include <windows.h>   // use as needed for your system
#include <gl\gl.h>  //Must include this before vectormatrix
#include <gl\glut.h>
#include <string.h>
#include <math.h> //For cos and sine functions
#include "timer.h"
#include "randomizer.h"
#include "vectormatrix.h"
#include "polyhedron.h" //For cAssembly and cPolyhedron
#include "myshapes.h" //For cCheerLeader
#include "texture.h"
#include <deque> //For "deque" template used for deques, which are double-ended queues
using namespace std; //Need this for standard template library to work.
#include "glext.h" /* This is for using OpenGL methods higher than
	those in the Version 1.1 that Windows supports.  I downloaded
	this file from SGI
	http://oss.sgi.com/projects/ogl-sample/ABI/glext.h
	and I found out about it at
	http://www.gamedev.net/reference/programming/features/oglext/
	after google searching for "Windows OpenGL 1.2" 
	The glext.h has a bunch of typedef that make it easy to 
	use OpenGL extensions.  We use the higher-than-1.1 functions
	as extensions. I initially got into this because I wanted to
	call the OpenGL 1.2 method glLightModeli(
		GL_LIGHT_MODEL_COLOR_CONTROL, GL_SEPARATE_SPECULAR_COLOR);
*/
#include "switches.h"
	/* The switches.h file contains some #define switches used
	by both donut.cpp and polyhedron.cpp.  In there are, for 
	instance, the #define for TESTTEXTURE and TESTALPHA. */
#ifdef TESTTEXTURE
	char * critterskinfile1 = 
		//"alien.bmp"; //Works.  I got this off the web.
		//"bignose.pcx";
		"mohawk.bmp"; //Works.  I made this in Windows paint.
			//"filigree.bmp"; //Doesn't work. I saved from PhotoShop.
	char * backgroundfile = 
			//"bignose.pcx";  //Works.  I got it off the web.
				"mohawk.bmp"; //Works.  I made this in Windows paint.
	//char * foregroundfile = "mohawk.bmp"; //"city.bmp";
	//char * foregroundfile = "city.bmp";
	char * foregroundfile = "city2.bmp";
	char * floorfile ="mohawk.bmp";//"bignose.pcx";
cTexture *pcrittertexture1, *pbackgroundtexture, *pforegroundtexture,
	*pfloortexture;
#endif // TESTTEXTURE

#define DONUTSTORM
	/* comment in DONUTSTORM to show a falling storm of polyhedra. */
//#define TESTCHEER
	/* comment in TESTCHEER to show display a gun turret cheerleader,
	also to move the eye to a different location. */

cPerformanceTimer timer; //A class I defined in timer.h
deque<cAssembly*> assemblyarray; //Seems faster and cleaner to use pointers.

GLdouble ZOOMRATIO = 0.8;
GLdouble startwindowwidth = 640.0;
GLdouble aspect = 4.0/3.0;
GLdouble xradius = 20.0;
GLdouble yradius = 15.0;
GLdouble zradius = 50.0;
#ifdef TESTCHEER
cVector eye(1.1 * xradius, 0.1 * yradius, 0.25 * zradius); //Better view of cheer
#else //not TESTCHEER
//cVector eye(0.0, - 0.05 * yradius, - 1.1 * zradius); //Used to be (0, -7.5, -55)
//cVector eye(0.0, 0.75 * yradius,  1.1 * zradius);// (0, 10, 55).
cVector eye(0.2 * xradius, 1.5 * yradius,  2.5 * zradius);// (0, 10, 55).
	/* In molecules 5 I turned things so the eye is in standard position at hiz
	looking towards loz. */
#endif //not TESTCHEER
double targetx = 0.0, targety = 0.0, targetz = 0.0;
double dtarget = 0.5;
double eyex = 0.0, eyey = 0.0, eyez = 0.0;
double deye = 0.5;
double zoom = 0.8;
GLdouble MINTIMEPERFRAME = (1.0/60.0); //Min time in seconds per frame
Real dt;
Real age = 0.0; //Time the program's been running.
bool animateflag = true;
bool fallflag = true;
bool threedflag = true;
bool lightingflag = true; 
//Draw flags. These are set to bit values defined in polyhedron.cpp
int showconnectorflag = 0;//cAssembly::SHOWCONNECTORFLAG; //Can toggle with 'c'
int showradiusflag = 0; //= cAssembly::SHOWRADIUSFLAG; //Can toggle with 'b'
int wireflag = 0; //= cAssembly::WIREFRAMEFLAG; //Can toggle with 'w'
#ifdef DONUTSTORM
	int maxassemblycomplexity = 0; //Can change with 'd/D'
	int maxassemblycount = 35; //Number of polyhedra to have in the storm.
	Real FALLSPEED = 2.0*zradius / 8.0; /* Means it should take 16.0 seconds
	to fall from hiz to loz. */
#else //not DONUTSTORM means Showing assemblies.
	int maxassemblycomplexity = 2; /* 	maxassemblycomplexity measures
		how many branchings deep the polyhedra are. */
	int maxassemblycount = 6; //Number of polyhedra to have in the storm.
	Real FALLSPEED = 2.0*zradius / 24.0; /* Means it should take denominator
		seconds to fall from hiz to loz. */
#endif //DONUTSTORM
#define BLACKBACKGROUND
#define ANTIALIAS

//#define TESTDEMO
#ifdef TESTDEMO
bool demomode =true;
char *demostring = "MMMh....q";
int demostringlength = 9;
#else //not TESTDEMO
bool demomode = false;
char *demostring = NULL;
int demostringlength = 0;
#endif //TESTDE
int demostringindex = 0;
Real nextdemokeyage = 0.0; //Time to get next char from demostring if demomode true.

void myLighting(bool onoff); //Body of this function is defined below.

//<<<<<<<<<<<<<<<<<<<<<<< assemblyarray >>>>>>>>>>>>>>>>>>>>
void emptyAssemblyArray()
{
	for (int i=0; i<assemblyarray.size(); i++)
			delete assemblyarray[i]; //Delete it before removing from the deque.
	assemblyarray.clear(); /* Removes all members by calling erase for each
		member.  The erase call doesn't call the delete method, it only
		destroys the four bytes of space that was occupied by the pointer,
		which is why we first walk the loop and do the deletes. */
}
//<<<<<<<<<<<<<<<<<<<<<<< myInit >>>>>>>>>>>>>>>>>>>>

void setProjectionMatrix(double zoom)
{
	glMatrixMode(GL_PROJECTION); 
	glLoadIdentity();
	if (!threedflag) //Use a 2D ortho projection.
		glOrtho(-zoom *xradius, zoom *xradius, -zoom *yradius, zoom *yradius,
			 -zradius, zradius); 
		/* Arguments: left, right, bottom, top, near, far.  The near and far are
		relative to a viewer located at the origin, so here we normally
		want a negative near and a positive far. We zoom only in the z and y,
		as zooming in the z might clip unnecessariy. */
	else //Use a 3D perspective projection
	{
		/*First I'd like to figure out how distant my far clip plane should be
			Really I ought to test all eight corners of the world box, but
			for now I'm just testing the obvious hi and lo corners. */
		double testdistance, farclip = 0.0;
		cVector testcorner;
		//Test corners
		for (int xs = -1; xs <= 1; xs +=2)
			for (int ys = -1; ys <= 1; ys +=2)
				for (int zs = -1; zs <= 1; zs +=2)
				{
					testcorner.set(xs*xradius, ys*yradius, zs*zradius);
					testdistance = (testcorner - eye).magnitude();
					if (testdistance > farclip)
						farclip = testdistance;
				}
		gluPerspective(zoom*45.0, aspect, 0.1, farclip);
		/* Arguments: view angle in degrees, width-to-height aspect ratio, 
		near, far.  The aspect should match the viewport dx/dy.  The near
		and far should be both positive, as here we think of the eye as
		standing back from the scene. */
		glMatrixMode(GL_MODELVIEW); /* In Molecules 6 I learned from David Lin
			that I need to have the gluLookAt acting on the MODELVIEW matrix
			and not on the PROJECTION matrix.  It acts on whatever matrix is
			active, as the OPenGL doc says, "The matrix generated by gluLookAt 
			postmultiplies the current matrix."  Not having this line here
			meant that our viewpoint calcuations were messed up and specular
			highlights were in the wrong place when we put in
			::glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE ); */
		glLoadIdentity(); /* Clear out any old gluLookAt resuts from 
			the MODELVIEW matrix. */
		gluLookAt(eye.x(), eye.y(), eye.z(),
			targetx, targety, targetz,
			0.0, 1.0, 0.0);
		/* Arguments: 3 eye coords, 3 center coords, 3 up direction components.
		This puts the eye at the orgin of a viewer coordinate system and puts
		the "center" on the negative z axis. Up until molcules 5, I had a
		bug here in that I had the up parameters as (0, 0, 1), the z axis,
		which was virtually the same as the look at direction, so the 
		confused gluLookAt was in fact adopting negative Y-axis as the up.*/
	}
}

void myInit(void)
{
/* We call myInit once at startup. */
#ifdef BLACKBACKGROUND
    glClearColor(0.0, 0.0, 0.0, 0.0);       // set black background color
#else //not BLACKBACKGROUND
    glClearColor(1.0, 1.0, 1.0, 0.0);       // set white background color
#endif //BLACKBACKGROUND
	setProjectionMatrix(1.0);
	glEnable(GL_LINE_SMOOTH);
	//glShadeModel(GL_FLAT);
	glShadeModel(GL_SMOOTH);
	glEnable(GL_DEPTH_TEST); //So I can use 3D
	myLighting(lightingflag);
	timer.setMinDt(MINTIMEPERFRAME); //Don't run faster than this rate

#ifdef TESTTEXTURE
	::glEnable(GL_TEXTURE_2D);
	 /* For the lighting calculations
		on the meshes, I'll treat the creatutures as as white.  I'll use
		the specular component of the lighting later, thanks to the fact
		that I've called ::glLightModeli(GL_LIGHT_MODEL_COLOR_CONTROL,
		GL_SEPARATE_SPECULAR_COLOR) in my myLighting() method; */
//::glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
::glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, 
		GL_MODULATE);
//		GL_REPLACE);
::glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
	//Hints can be GL_NICEST, GL_FASTEST, or GL_DONT_CARE
	//Perspective textures look better if done with hyperbolic 
	//interpolation instead of linear interpolation
#endif //TESTTEXTURE

#ifdef TESTALPHA
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
#endif //TESTALPHA	

#ifdef TESTFOG
	int fogMode = GL_EXP2;
	glEnable(GL_FOG);
	glFogi(GL_FOG_MODE, fogMode);
	GLfloat redfog[4] = {1.0, 0.0, 0.1, 1.0};
	glFogfv(GL_FOG_COLOR, redfog);
	glFogf(GL_FOG_DENSITY, 0.03); // used for GL_EXP & GL_EXP2
	glFogf(GL_FOG_START, 0.01); // start/end is used for GL_LINEAR
	glFogf(GL_FOG_END, 0.02);
#endif //TESTFOG

#ifdef TESTBILLBOARD
				/*	We are going to set the transparent parts of our textures to have
				alpha value of zero.  There are two possible ways to let things "show through"
				the transparent alpha.  We presently do a clipping-style alpha test.  Even if
				we were to turn on blending, some tests indicate that without the alpha test,
				the textures would only be transparent in one direction. 
					This technique for getting a transparent background of a texture is
				mentioned in the Red Book, and indexed under "billboarding". We have set
				the alpha to 0 for our transparent texture backgrounds, but we need
				the GL_ALPHA_TEST as well, otherwise the "transparent" pixels can still cover
				up underlying pixels (unless we enable GL_BLEND, but I find the GL_ALPHA_TEST
				easier to get working, as described in the next comment. */
			::glEnable(GL_ALPHA_TEST);
			::glAlphaFunc(GL_GREATER, 0.99f);
#endif //TESTBILLBOARD
}

//<<<<<<<<<<<<<<<<<<<<<<< drawFlurry >>>>>>>>>>>>>>>>>
void drawFlurry() //Uses the global dt param.
{
	while (assemblyarray.size() < maxassemblycount)
	{
		cAssembly *pnewpoly;
#ifdef TESTCHEER
		pnewpoly = new cCheerleader();
#else	//not TESTCHEER
		pnewpoly = cAssembly::makeRandomAssembly(maxassemblycomplexity);
#endif //TESTCHEER
			pnewpoly->setPosition(cVector(cRandomizer::pinst()->randomReal(-0.5* xradius, 0.5 * xradius),
				cRandomizer::pinst()->randomReal(-0.5* yradius, 0.5* yradius),
				cRandomizer::pinst()->randomReal(-zradius, 0.0)));

		assemblyarray.push_back(pnewpoly); 
			//push_back adds a copy of the polygon to the array.
	}
	for (int i=0; i<assemblyarray.size(); i++)
	{
		assemblyarray[i]->animate(dt);
		if (fallflag)
			assemblyarray[i]->translate(FALLSPEED * dt * cVector::ZAXIS);
#ifdef TESTTEXTURE
		//cPolyhedron *phed = (cPolyhedron *)(assemblyarray[i]);
		//if (phed && phed->usestexture()) //Cast so I can call usestexture
			pcrittertexture1->select(); //I allocated this texture in the main method.
#endif //TESTTEXTURE
		assemblyarray[i]->draw(wireflag | showconnectorflag | showradiusflag);
		if (assemblyarray[i]->position().z() > 
			zradius - assemblyarray[i]->radius()) 
		{
			delete assemblyarray[i]; //Delete it before removing from the deque.
			assemblyarray.erase(assemblyarray.begin()+i); //Erase when you hit bottom.
		}
	}
}  
//<<<<<<<<<<<<<<<<<<<<<<<< myDisplay >>>>>>>>>>>>>>>>>
void myDisplay(void)
{
/* myDisplay is called once at start up and is called again whenever
the graphics window is first covered and then uncovered, making a
redraw necessary. */
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);     // clear the screen 
//Draw 3D box around the world.  Disable lighting for lines.
	GLboolean texturewason, lightingwason;
	::glGetBooleanv(GL_TEXTURE_2D, &texturewason);
	::glGetBooleanv(GL_LIGHTING, &lightingwason);
#ifdef TESTTEXTURE
	setMaterialColor(cVector(1.0, 1.0, 1.0));
	pbackgroundtexture->select(); //I allocated this texture in the main method.
	glBegin(GL_QUADS); //loz face, back wall
		glNormal3f(0.0, 0.0, 1.0);
		glTexCoord2f(0.0, 0.0); glVertex3d(-xradius, -yradius, -zradius);
		glTexCoord2f(1.0, 0.0); glVertex3d(xradius, -yradius, -zradius);
		glTexCoord2f(1.0, 1.0); glVertex3d(xradius, yradius, -zradius);
		glTexCoord2f(0.0, 1.0); glVertex3d(-xradius, yradius, -zradius);
	glEnd();
	pforegroundtexture->select(); //I allocated this texture in the main method.
	glBegin(GL_QUADS); //loz face, front wall
		glNormal3f(0.0, 0.0, 1.0);
		glTexCoord2f(0.0, 0.0); glVertex3d(-xradius, -yradius, zradius);
		glTexCoord2f(1.0, 0.0); glVertex3d(xradius, -yradius, zradius);
		glTexCoord2f(1.0, 1.0); glVertex3d(xradius, yradius, zradius);
		glTexCoord2f(0.0, 1.0); glVertex3d(-xradius, yradius, zradius);
	glEnd();
	pfloortexture->select(); //I allocated this texture in the main method.
	//glDisable(GL_TEXTURE_2D);
			 //loyface, floor
	glBegin(GL_QUADS); //loy face, floor
		glNormal3f(0.0, 1.0, 0.0);
		glTexCoord2f(0.0, 0.0); glVertex3d(-xradius, -yradius, zradius);
		glTexCoord2f(1.0, 0.0); glVertex3d(xradius, -yradius, zradius);
		glTexCoord2f(1.0, 1.0); glVertex3d(xradius, -yradius, -zradius);
		glTexCoord2f(0.0, 1.0); glVertex3d(-xradius, -yradius, -zradius);
	glEnd();
#endif //TESTTEXTURE
//Draw lines
	setVertexColor(cVector(0.8, 0.8, 0.8)); //Light gray
	glLineWidth(2.0);
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_LIGHTING);
	Real eo = 0.1; //Use eo as an edge offset to prevent zfighting.
	glBegin(GL_LINE_LOOP); //lozface
		glVertex3d(-xradius+eo, -yradius+eo, -zradius+eo);
		glVertex3d( xradius-eo, -yradius+eo, -zradius+eo);
		glVertex3d( xradius-eo,  yradius-eo, -zradius+eo);
		glVertex3d(-xradius+eo,  yradius-eo, -zradius+eo);
	glEnd();
	glBegin(GL_LINE_LOOP); //hizface
		glVertex3d(-xradius+eo, -yradius+eo, zradius-eo);
		glVertex3d( xradius-eo, -yradius+eo, zradius-eo);
		glVertex3d( xradius-eo,  yradius-eo, zradius-eo);
		glVertex3d(-xradius+eo,  yradius-eo, zradius-eo);
	glEnd();
	glBegin(GL_LINE_LOOP); //loyface
		glVertex3d(-xradius+eo, -yradius+eo,  zradius-eo);
		glVertex3d( xradius-eo, -yradius+eo,  zradius-eo);
		glVertex3d( xradius-eo, -yradius+eo, -zradius+eo);
		glVertex3d(-xradius+eo, -yradius+eo, -zradius+eo);
	glEnd();
	glBegin(GL_LINE_LOOP); //hiyface
		glVertex3d(-xradius+eo, yradius-eo,  zradius-eo);
		glVertex3d( xradius-eo, yradius-eo,  zradius-eo);
		glVertex3d( xradius-eo, yradius-eo, -zradius+eo);
		glVertex3d(-xradius+eo, yradius-eo, -zradius+eo);
	glEnd();
	glBegin(GL_LINE_LOOP); //loxface
		glVertex3d(-xradius+eo, -yradius+eo,  zradius-eo);
		glVertex3d(-xradius+eo, -yradius+eo, -zradius+eo);
		glVertex3d(-xradius+eo,  yradius-eo, -zradius+eo);
		glVertex3d(-xradius+eo,  yradius-eo,  zradius-eo);
	glEnd();
	glBegin(GL_LINE_LOOP); //hixface
		glVertex3d(xradius-eo, -yradius+eo,  zradius-eo);
		glVertex3d(xradius-eo, -yradius+eo, -zradius+eo);
		glVertex3d(xradius-eo,  yradius-eo, -zradius+eo);
		glVertex3d(xradius-eo,  yradius-eo,  zradius-eo);
	glEnd();
	//Draw coordinate axes at origin in Red, Blue, Green
	glLineWidth(3.0);
	glBegin(GL_LINES);
		setVertexColor(cVector(1.0, 0.0, 0.0)); //X axis red
		glVertex3d(0.0, 0.0, 0.0);
		glVertex3d(1.0, 0.0, 0.0);
		setVertexColor(cVector(0.0, 0.0, 1.0));//Y axis blue
		glVertex3d(0.0, 0.0, 0.0);
		glVertex3d(0.0, 1.0, 0.0);
		setVertexColor(cVector(0.0, 1.0, 0.0)); //Z axis Green
		glVertex3d(0.0, 0.0, 0.0);
		glVertex3d(0.0, 0.0, 1.0);
	glEnd();
	//restore after line drawing.
	if (texturewason)	
		::glEnable(GL_TEXTURE_2D);
	if (lightingwason)
		::glEnable(GL_LIGHTING);
	//Draw polygons
	drawFlurry();
	glFlush();				// Single buffer: send all output to display 
	glutSwapBuffers();		//Double buffer: flip.
}
//<<<<<<<<<<<<<<<<<<<<<<<< myReshape >>>>>>>>>>>>>>>>>

void myReshape(int width, int height)
{
	aspect = double(width)/double(height);
	setProjectionMatrix(zoom);
	glViewport(0, 0, width, height);
}
//<<<<<<<<<<<<<<<<<<<<<<<< myKeyboard >>>>>>>>>>>>>>>>>
void myKeyboard(unsigned char key, int mx, int my)
{
/* key is the ASCII code of the key, such as 'a', 'A', 'b', 'B', and
so on.  Note that case matters.  mx and my are the current mouse
position. */
	switch(key)
	{
		case ' ': //Space
			myDisplay(); //redraw 
			break;
		case 'b':
			showradiusflag ^= cAssembly::SHOWRADIUSFLAG;
			break;
		case 'c':
			showconnectorflag ^= cAssembly::SHOWCONNECTORFLAG;
			break;
		case 'f':
			animateflag ^= true;
			break;
		case 'h':
			fallflag ^= TRUE;
			break;
		case 'm':
			maxassemblycomplexity--;
			if (maxassemblycomplexity < 0)
				maxassemblycomplexity = 0;
			emptyAssemblyArray();
			break;
		case 'M':
			maxassemblycomplexity++;
			emptyAssemblyArray();
			break;
		case 'n':
			emptyAssemblyArray();
			maxassemblycount--;
			if (maxassemblycount < 1)
				maxassemblycount = 1;
			break;
		case 'N':
			emptyAssemblyArray();
			maxassemblycount++;
			if (maxassemblycount < 1)
				maxassemblycount = 1;
			break;
		case 'p':
			threedflag ^= true;
			setProjectionMatrix(zoom);
			break;
		case 'q':
			exit(-1); //Terminate
			break;
		case 'r':
			emptyAssemblyArray(); 
			break;
		case 'A':
			targetx -= dtarget;
			setProjectionMatrix(zoom);
			break;
		case 'D':
			targetx += dtarget;
			setProjectionMatrix(zoom);
			break;
		case 'S':
			targety -= dtarget;
			setProjectionMatrix(zoom);
			break;
		case 'W':
			targety += dtarget;
			setProjectionMatrix(zoom);
			break;
		case 'I':
			targetz -= dtarget;
			setProjectionMatrix(zoom);
			break;
		case 'O':
			targetz += dtarget;
			setProjectionMatrix(zoom);
			break;
		case 'a':
			eyex = eye.x();
			eyex -= deye;
			eye.set(eyex, eye.y(), eye.z());
			setProjectionMatrix(zoom);
			break;
		case 'd':
			eyex = eye.x();
			eyex += deye;
			eye.set(eyex, eye.y(), eye.z());
			setProjectionMatrix(zoom);
			break;
		case 's':
			eyey = eye.y();
			eyey -= deye;
			eye.set(eye.x(), eyey, eye.z());
			setProjectionMatrix(zoom);
			break;
		case 'w':
			eyey = eye.y();
			eyey += deye;
			eye.set(eye.x(), eyey, eye.z());
			setProjectionMatrix(zoom);
			break;
		case 'i':
			eyez = eye.z();
			eyez -= deye;
			eye.set(eyex, eye.y(), eyez);
			setProjectionMatrix(zoom);
			break;
		case 'o':
			eyez = eye.z();
			eyez += deye;
			eye.set(eyex, eye.y(), eyez);
			setProjectionMatrix(zoom);
			break;
		case 'g':
			wireflag ^= cAssembly::WIREFRAMEFLAG;
			break;
		case 'z':
			zoom *= ZOOMRATIO;
			setProjectionMatrix(zoom);
			break;
		case 'Z':
			zoom *= 1.0/ZOOMRATIO;
			setProjectionMatrix(zoom);
			break;
	}
}
//<<<<<<<<<<<<<<<<<<<<<<<< myMouse >>>>>>>>>>>>>>>>>
void myMouse(int button, int state, int mx, int my)
{
/* Gets called when a button is clicked.  The arguments are:
button is GLUT_LEFT_BUTTON or GLUT_RIGHT_BUTTON,
state is GLUT_DOWN or GLUT_UP, 
and mx and my are window pixel coordinates, origin at window upper left. */
}

//<<<<<<<<<<<<<<<<<<<<<<<< mMovedMouse >>>>>>>>>>>>>>>>>
void myMovedMouse(int mx, int my) //click and drag
//void myPassiveMovedMouse(int mx, int my) //Just drag
{
/*Gets called whenever the mouse moves, and mx and my are the new window
pixel coordinates, origin at window upper left. */

	static int oldmx=0, oldmy = 0;
	static bool firsttime = true;

	if (firsttime)
	{
		oldmx = mx;
		oldmy = my;
		firsttime = false;
		return;
	}
	if (mx < oldmx)
	{
		targetx -= dtarget;
		setProjectionMatrix(zoom);
	}
	if (mx > oldmx)
	{
		targetx += dtarget;
		setProjectionMatrix(zoom);
	}
	if (my < oldmy)
	{
		targety += dtarget;
		setProjectionMatrix(zoom);
	}
	if (my > oldmy)
	{
		targety -= dtarget;
		setProjectionMatrix(zoom);
	}
	oldmx = mx;
	oldmy = my;
}


//<<<<<<<<<<<<<<<<<<<<<<<< myAnimate >>>>>>>>>>>>>>>>>
void myAnimate(void)
{
/* myAnimate is called over and over whenever glut is idle (not busy
processing a key or mouse message). To keep this from happening too
often we put in a delay, that is, we use a spin loop to make myAnimate
take at least MINTIMEPERFRAME seconds. */
	if (!animateflag)
		return;
	dt = timer.tick();
	age += dt;
	if (demomode && age >= nextdemokeyage)
	{
		nextdemokeyage = age + 0.1;
		char demokey = demostring[demostringindex];
		demostringindex++;
		if (demostringindex >= demostringlength)
			demostringindex = 0;
		if (demokey == '.')
			nextdemokeyage += 1.0;
		else if (demokey == ',')
			nextdemokeyage += 0.25;
		else
			myKeyboard(demokey, 0, 0); //params 2 and 3 are mouse coords, use dummy for now
	}
	myDisplay();
}
//<<<<<<<<<<<<<<<<<<<<<<<< myLighting >>>>>>>>>>>>>>>>>
void myLighting(bool onoff)
{
	if (onoff)
		::glEnable(GL_LIGHTING);
	else
	{
		::glDisable(GL_LIGHTING);
		return;
	}
//(Step 1) Define and install the material properties for objects. =================
	/* These are just defaults.  In practice we can alter all of these
		material properties with calls inside each of our graphic objects'
		draw methods, thus giving each object its own intrinsic material.
		In this demo, we will in fact be resetting the ambient and diffuse
		color for each cPolyhedron. */ 
	GLfloat materialSpecular[4] = { 1.0f, 1.0f, 1.0f, 1.0f };
		/* Normally a material will reflect a specular highlight in the
		same color as the incoming specular light, so we usually pick
		a shade of gray here. If you make these numbers too high and have 
		a broad highlight you can get an unpleasant "popping" effect where
		a facet flares up into high illumination all at once. */
	GLfloat materialShininess[1] = { 70.0f }; 
		/* Bigger means smaller sized highlight.  Range is 0.0 to 128.0.
		OpenGL insists that you have to place this real inside
		a single element array. */
	::glMaterialfv( GL_FRONT, GL_SPECULAR, materialSpecular );
	::glMaterialfv( GL_FRONT, GL_SHININESS, materialShininess );
	GLfloat whiteAmbDiff[4]={ 1.0, 1.0, 1.0, 1.0f };
	::glMaterialfv( GL_FRONT, GL_AMBIENT_AND_DIFFUSE, 
		whiteAmbDiff );  //Just to put in a default
		/* We commonly use the same material color for both ambient and
		diffuse lighting.  Our cPolyhedron::draw code will in fact
		call glMaterialfv to give each polyhedron a color.   Note that the 
		::glColorf4(r,g,b,alpha) call only works when lighting is OFF, and
		has no effect at all when lighting is ON.  glMaterialfv(...) on the
		other hand only works when lighting is ON, and has no effect when
		lighting is OFF. */

//(Step 2)Define and install the LightModel settings. ======================. 
//	GLfloat global_ambient[4] = { 0.75f, 0.75f, 0.75f, 1.0f }; //fairly bright
//	::glLightModelfv(GL_LIGHT_MODEL_AMBIENT, global_ambient);
//Don't use global ambient, it washes things out.
	::glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE );
		/* If I enable GL_LIGHT_MODEL_LOCAL_VIEWER, it runs a bit slower but
		the highlights are more accurate for moving eye.  By default
		this  is turned off and the calculation assumes the viewer to is
		looking roughly	the negative z direction, which is in fact the case
		in this demo, so turning this on won't change much. But if you were
		to have	GL_LIGHT_MODEL_LOCAL_VIEWER off and put the eye down at the 		negative z and look in the positive z direction, then the specular highlights
		would be totally wrong. */
//	::glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE );
		/* If I use the GL_LIGHT_MODEL_TWO_SIDE, I am allowed to use
		different material on the back faces of objects. Otherwise, I think OpenGL
		by default uses the same material as is on the front of the face.  In
		any case, back faces are rendered by taking the front surface normal
		and flipping it so as to get a normal to the back face.  I think this
		switch only makes a difference if I make some
		 ::glMaterialfv( GL_BACK...)calls as well as
		::glMaterialfv( GL_FRONT...) calls.  */
	::glLightModeli(GL_LIGHT_MODEL_COLOR_CONTROL, 
		GL_SEPARATE_SPECULAR_COLOR);
		/* This call is an OpenGL 1.2 method, and I need to include
		glext.h for it to work, as wretched Windows still ships with only
		OpenGl 1.1, trying to make us use DirectX.  Thanks, Bill... 
		Anyway, the point of the call is to put nice specular highlights
		on top of textured objects. To turn this off, call with
		GL_SINGLE_COLOR in the second arg. */

//(3a) Define the individual lights' properties. ===============================
/* The ambient is a directionless light that's simply added in;  since we have
a global_ambient above it's less confusing to have the individual light
ambients be zero.
	The diffuse is closest to the "color" of the light.
	The specular is used for color highlights, logically it can match diffuse,
 but it looks peppier if its different.
	We have a DIRECTIONAL light if the position has 0 in the fourth coord.
In this case (x,y,z) describes the direction towards the light, which means 
that the light is treated as parallel lines from infinity going in the
direction from (x,y,z) towards (0,0,0), in other words, going in the direction
(-x, -y, -z).  Yes, this is confusing.
	We have a POSITIONAL light if we put the position has a 1 in the fourth
coordinate.  In this case (x,y,z) is the location of the light, which radiates
in all directions form this point.  If (x,y,z) is way out past your eye, then
(x,y,z,1) acts very much like (x,y,z,0).  BUT if (x,y,z) is closer to the origin
than your eye position, the effect is reversed, that is, the effective direction
of the light rays for you will be in the (x,y,z) direction. */

/* Let's try four lights, and give them different colors for the different
directions, for the sake of visual variety.  One goal would be to pick the
lights so things are nicely illuminated no matter where you put the eye,
another approach would be to tailor the lights to look good relative to
our current eye position of (0, 10, 55) --- though note that this position
changes if I use TESCHEER.  It might be a good idea to
and make the net sum of the various colored lights be nearly white. We won't
use ambient for the sources as we do that in the lighting model. */

	GLfloat ambient0[] =  { 0.0f, 0.0f, 0.0f, 0.0f }; //We have global ambient.
	GLfloat diffuse0[] =  { 0.0f, 0.0f, 0.6f, 1.0f };//Blue
	GLfloat specular0[] = { 0.0f, 0.0f, 0.6f, 1.0f };
	GLfloat position0[] = { 10.0f, 50.0f, 20.0f, 0.0f }; //Light towards -x,-y,-z
		/* If I put 0 in the last term of position I get a directional light
		that comes from infinity in the direction gotten by placing minuses in
		front of the first three numbers. If I am putting my eye out at 
		positive values, then aiming towards negative values is  a good idea
		as this effectively
		puts the light come "over my shoulder". If I put in negative
		"position" values for a directional light the light is at infinity
		shining in the positive direction, so it will only light up the 
		"back" or negative sides of objects, and if my eye is out at the positive
		side, then I won't see the light.   */

	GLfloat ambient1[] =  { 0.0f, 0.0f, 0.0f, 0.0f };//We have global ambient..
	GLfloat diffuse1[] =  { 0.6f, 0.1f, 0.1f, 1.0f }; //Red
	GLfloat specular1[] = { 0.6f, 0.1f, 0.1f, 1.0f };
	GLfloat position1[] = { -100.0f, 20, 65, 1.0f }; /* My eye is at (0,10,55)
			so this is effectively towards x,-y,-z. */
		/* If I put 1 in the last term of position I get a light
		located at the first three numbers and glowing in every direction
		from that location. I should probably place my positional lights at 
		positive values as well perhaps at values even higher than my eye
		coordinates.  Because if I put the light down at the negative values,
		then the light	will just be lighting up the invisible side of the 
		object. Note that the situation is just the opposite when I use a
		directional light (last coordinate 0) instead of a positional
		light (last coordinate 0). */

	GLfloat ambient2[] =  { 0.0f, 0.0f, 0.0f, 0.0f };//We have global ambient.
	GLfloat diffuse2[] =  { 0.4f, 0.6f, 0.1f, 1.0f }; //Yellow
	GLfloat specular2[] = { 0.4f, 0.6f, 0.1f, 1.0f };
	GLfloat position2[] = { -10.0f, 0.0f, 1.0f, 0.0f }; //Towards x, 0, -z
 
	GLfloat ambient3[] =  { 0.0f, 0.0f, 0.0f, 0.0f };//We have global ambient.
	GLfloat diffuse3[] =  { 1.0f, 1.0f, 1.0f, 1.0f }; //white
	GLfloat specular3[] = { 0.55f, 0.650f, 0.55f, 1.0f };
	GLfloat position3[] = { 0.0f, 0.0f, 70.0f, 1.0f }; /* Above and Behind me. */

//(3b) Enable your lights and Install their properties =========================
//Enable and install Light0
	::glEnable(GL_LIGHT0);  
	::glLightfv(GL_LIGHT0, GL_AMBIENT, ambient0);
	::glLightfv(GL_LIGHT0, GL_POSITION, position0);
	::glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuse0);
	::glLightfv(GL_LIGHT0, GL_SPECULAR, specular0);
//Enable and install Light1
	::glEnable(GL_LIGHT1);
	::glLightfv(GL_LIGHT1, GL_AMBIENT, ambient1);
	::glLightfv(GL_LIGHT1, GL_POSITION, position1);
	::glLightfv(GL_LIGHT1, GL_DIFFUSE, diffuse1);
	::glLightfv(GL_LIGHT1, GL_SPECULAR, specular1);
//Enable and install Light2
	::glEnable(GL_LIGHT2);
	::glLightfv(GL_LIGHT2, GL_AMBIENT, ambient2);
	::glLightfv(GL_LIGHT2, GL_POSITION, position2);
	::glLightfv(GL_LIGHT2, GL_DIFFUSE, diffuse2);
	::glLightfv(GL_LIGHT2, GL_SPECULAR, specular2);
//Enable and install Light3 
	::glEnable(GL_LIGHT3);
	::glLightfv(GL_LIGHT3, GL_AMBIENT, ambient3);
	::glLightfv(GL_LIGHT3, GL_POSITION, position3);
	::glLightfv(GL_LIGHT3, GL_DIFFUSE, diffuse3);
	::glLightfv(GL_LIGHT3, GL_SPECULAR, specular3);

/* For testing, I can try disabling all but one of the lights by commenting in any three of the following lines. */
	//::glDisable(GL_LIGHT0);  
	//::glDisable(GL_LIGHT1);  
	//::glDisable(GL_LIGHT2);  
	//::glDisable(GL_LIGHT3);  

}

//<<<<<<<<<<<<<<<<<<<<<<<< main >>>>>>>>>>>>>>>>>>>>>>
void main(int argc, char** argv)
{
	if (argc > 1)
	{
		demomode = true;
		demostring = argv[1];
		demostringlength = strlen(demostring);
	}
	glutInit(&argc, argv);          // initialize the toolkit
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB ); // set display mode
	glutInitWindowSize(startwindowwidth, int(startwindowwidth/aspect));     // set window size
	glutInitWindowPosition(100, 150); // set window position on screen
	glutCreateWindow("Texture 6, CS 116B, May 18, 2004, Rudy Rucker, His Last Demo."); // open the screen window
	glutDisplayFunc(myDisplay);     // register redraw function
	glutReshapeFunc(myReshape);
	glutKeyboardFunc(myKeyboard);	// register keyboard function.
	glutMouseFunc(myMouse);			// register mouse click function.
	glutMotionFunc(myMovedMouse);   // register mouse move function.
	//glutPassiveMotionFunc(myPassiveMovedMouse);   // register mouse move function.
	glutIdleFunc(myAnimate);		// register idle function to call.
#ifdef TESTTEXTURE
/* In the cTexture constructors, the second argument specifies if you
want the texture to have transparent background. */
	pcrittertexture1 = new cTexture(critterskinfile1, false);
	pbackgroundtexture = new cTexture(backgroundfile, true);
	pforegroundtexture = new cTexture(foregroundfile, true);
	pfloortexture = new cTexture(floorfile, false);
#endif //TESTTEXTURE
	myInit();                   
	glutMainLoop(); 		     // go into a perpetual loop
#ifdef TESTTEXTURE
	delete pcrittertexture1;
	delete pbackgroundtexture;
	delete pforegroundtexture;
	delete pfloortexture;
#endif //TESTTEXTURE
}
